package org.netkernel.ext.system.accessor;

import java.io.BufferedReader;
import java.io.File;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.lang.management.GarbageCollectorMXBean;
import java.lang.management.ManagementFactory;
import java.lang.management.MemoryMXBean;
import java.lang.management.MemoryPoolMXBean;
import java.lang.management.MemoryUsage;
import java.lang.management.ThreadInfo;
import java.lang.management.ThreadMXBean;
import java.net.InetAddress;
import java.net.JarURLConnection;
import java.net.URI;
import java.net.URL;
import java.net.URLClassLoader;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.Map.Entry;
import java.util.jar.Attributes;
import java.util.jar.Manifest;

import org.netkernel.container.IKernel;
import org.netkernel.container.config.IConfiguration;
import org.netkernel.container.config.IConfigurationListener;
import org.netkernel.container.impl.Kernel;
import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.layer0.boot.IModule;
import org.netkernel.layer0.boot.ModuleManager;
import org.netkernel.layer0.boot.ModuleSourceWithLevel;
import org.netkernel.layer0.meta.IModuleMeta;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.GoldenThreadExpiryFunction;
import org.netkernel.layer0.util.Utils;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.request.IRequest;
import org.netkernel.request.impl.RequestFactory;
import org.netkernel.scheduler.GroundState;
import org.netkernel.scheduler.IRequestState;
import org.netkernel.scheduler.PriorityRequestTable;
import org.netkernel.scheduler.RequestState;
import org.netkernel.scheduler.RootState;
import org.netkernel.scheduler.Scheduler;
import org.netkernel.scheduler.WorkerThread;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpaceWithIdentity;
import org.netkernel.urii.impl.NetKernelException;
import org.netkernel.util.MultiMap;

public class KernelProcAccessor extends StandardAccessorImpl implements IConfigurationListener
{	
	private SystemStatisticsMonitor mStatisticsMonitor;
	private GoldenThreadExpiryFunction mConfigurationExpiry;
	
	public KernelProcAccessor()
	{	this.declareThreadSafe();
		mConfigurationExpiry=new GoldenThreadExpiryFunction("IConfigurationListener");
	}

	@Override
	protected void postCommission(INKFRequestContext context) throws Exception
	{
		super.postCommission(context);
		mStatisticsMonitor=new SystemStatisticsMonitor();
		mStatisticsMonitor.init(context);
		IKernel k=(Kernel)context.getKernelContext().getKernel();
		k.addConfigurationListener(this);
	}
	
	@Override
	protected void preDecommission(INKFRequestContext context) throws Exception
	{	
		IKernel k=(Kernel)context.getKernelContext().getKernel();
		k.removeConfigurationListener(this);

		super.preDecommission(context);
		if (mStatisticsMonitor!=null)
		{	mStatisticsMonitor.shutdown();
		}
	}
	
	public void configurationChanged(IConfiguration aConfiguration)
	{	mStatisticsMonitor.configurationChanged(aConfiguration);
		mConfigurationExpiry.invalidate();
		mConfigurationExpiry=new GoldenThreadExpiryFunction("IConfigurationListener");
	}


	@Override
	public void onSource(INKFRequestContext context) throws Exception
	{	//Object result=null;
		//boolean shouldExpire=false;
		INKFResponse resp;
		int args=context.getThisRequest().getArgumentCount();
		if(args==1)
		{	HDSBuilder b=new HDSBuilder();
			b.pushNode("netkernel", null);
				b.addNode("runlevel", getRunLevel());
				b.addNode("runlevelmax", getMaxRunLevel());
				b.importNode(getUptime(context).getFirstNode("uptime"));
				b.importNode(getConfigSet(context).getFirstNode("config"));
				b.importNode(getThreads(context).getFirstNode("k"));
			b.popNode();
			
			resp=context.createResponseFrom(b.getRoot());
			resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
		}
		else
		{	String path1=context.getThisRequest().getArgumentValue(1);
			if(path1.equals("config"))
			{	if(args==2)
				{	
					resp=context.createResponseFrom(getConfigSet(context));
					resp.setExpiry(INKFResponse.EXPIRY_MIN_FUNCTION_DEPENDENT, mConfigurationExpiry);
				}
				else
				{	IConfiguration config=context.getKernelContext().getKernel().getConfiguration();
					try
					{	resp=context.createResponseFrom(config.getString(context.getThisRequest().getArgumentValue(2)));
						resp.setExpiry(INKFResponse.EXPIRY_MIN_FUNCTION_DEPENDENT, mConfigurationExpiry);
					}
					catch(Exception e) {/* Do nothing*/}		
				}
			}
			else if(path1.equals("uptime"))
			{	resp=context.createResponseFrom(getUptime(context));
				resp.setExpiry(INKFResponse.EXPIRY_CONSTANT,System.currentTimeMillis()+30000L);
			}
			else if(path1.equals("memory"))
			{	resp=context.createResponseFrom(mStatisticsMonitor.getBuffer());
				long statperiod=context.getKernelContext().getKernel().getConfiguration().getLong("netkernel.statistics.period");
				resp.setExpiry(INKFResponse.EXPIRY_CONSTANT,System.currentTimeMillis()+statperiod);
			}
			/*
			else if(path1.equals("memorySummary"))
			{	resp=context.createResponseFrom(mStatisticsMonitor.getSummary());
				long statperiod=context.getKernelContext().getKernel().getConfiguration().getLong("netkernel.statistics.period");
				resp.setExpiry(INKFResponse.EXPIRY_CONSTANT,System.currentTimeMillis()+statperiod);
			}
			*/
			else if(path1.equals("runlevel"))
			{	resp=context.createResponseFrom(getRunLevel());
				resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
			}
			else if(path1.equals("runlevelmax"))
			{	resp=context.createResponseFrom(getMaxRunLevel());
				resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
			}
			else if(path1.equals("k"))
			{	resp=context.createResponseFrom(getThreads(context));
				resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
			}
			else if (path1.equals("spaces"))
			{	resp=context.createResponseFrom(getKernelSpaces(context));
				resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
			}
			else if(path1.equals("environment"))
			{	resp=context.createResponseFrom(getEnvironment(context));
				resp.setExpiry(INKFResponse.EXPIRY_ALWAYS);
			}
			else if(path1.equals("deployment"))
			{	resp=context.createResponseFrom(getDeployment(context));
				resp.setExpiry(INKFResponse.EXPIRY_NEVER);
			}
		}
	}
	
	@Override
	public void onSink(INKFRequestContext context) throws Exception
	{	int args=context.getThisRequest().getArgumentCount();
		if(args>1)
		{	String path1=context.getThisRequest().getArgumentValue(1);
			if(path1.equals("runlevel"))
			{	String levels=context.sourcePrimary(String.class);
				int level=Integer.parseInt(levels);
				ModuleManager.getSingleton().setRunLevel(level);
			}
		}
	}
	
	private String getRunLevel()
	{	return Integer.toString(ModuleManager.getSingleton().getRunLevel());
	}
	
	private String getMaxRunLevel()
	{	return Integer.toString(ModuleManager.getSingleton().getMaxRunLevel());
	}
	
	
	private IHDSNode getConfigSet(INKFRequestContext context) throws Exception
	{	IConfiguration config=context.getKernelContext().getKernel().getConfiguration();
		HDSBuilder b=new HDSBuilder();
		b.pushNode("config",null);
		Iterator<String> i=config.getKeys();
		while(i.hasNext())
		{	String key=i.next();
			b.pushNode("item");
			{	b.addNode("name", key);
				b.addNode("value", config.getString(key));
			}
			b.popNode();
		}
		b.popNode();
		return b.getRoot();
	}

	private IHDSNode getUptime(INKFRequestContext context) throws Exception
	{	IConfiguration config=context.getKernelContext().getKernel().getConfiguration();
		long end=System.currentTimeMillis();
		long uptime=end-Long.parseLong(config.getString("netkernel.boot.time"));
		HDSBuilder b=new HDSBuilder();
		Kernel k=(Kernel)context.getKernelContext().getKernel();
		List<RootState> states=k.getScheduler().getRootRequestStates();
		long requests=0;
		if(states.size()>0)
		{	requests=states.get(0).getId();
		}
		b.pushNode("uptime");
			b.addNode("totalMilliseconds", uptime);
			b.addNode("totalRootRequests", requests);
			b.addNode("displayString",Utils.formatPeriod(uptime,false));
		b.popNode();
		return b.getRoot();	
	}
		
	private synchronized IHDSNode getThreads(INKFRequestContext context) throws Exception
	{
		IKernel k=context.getKernelContext().getKernel();
		long runawayPeriod=k.getConfiguration().getLong("netkernel.scheduler.thread.timeout",1000L);
		Kernel kImpl=(Kernel)k;
		Scheduler scheduler = kImpl.getScheduler();
		MultiMap lockedEndpoints=scheduler.dumpBusyEndpoints();
		RequestState[] queue=scheduler.dumpRequestTable();
		List<RootState> rootStates = scheduler.getRootRequestStates();
		ThreadMXBean threadBean=java.lang.management.ManagementFactory.getThreadMXBean();
		
		//find threads and last activity associated with each root
		Map<Long,RequestState> lastStateOnThread = new HashMap<Long,RequestState>();
		Map<RootState, RequestState> lastStateOnRoot = new HashMap<RootState, RequestState>();
		for (RootState root: rootStates)
		{	RequestState latestForRoot=recurseStateForThreads(root,lastStateOnThread);
			lastStateOnRoot.put(root, latestForRoot);
		}
		
		
		//get threadinfo for all relevant threads
		List<WorkerThread> workerThreads=scheduler.getActiveThreads();
		List<Long> tidList=new ArrayList<Long>(workerThreads.size());
		for (WorkerThread wt : workerThreads)
		{	tidList.add(wt.getId());
		}
		for (Long tid: lastStateOnThread.keySet())
		{	if (!tidList.contains(tid))
			{	tidList.add(tid);
			}
		}
		/*
		long[] ids = new long[tidList.size()];
		for (int i=0; i<ids.length; i++)
		{	Long tid=tidList.get(i);
			ids[i]=tid;
		}
		*/
		
		long[] ids=threadBean.getAllThreadIds();
		
		
		int maxDepth=NetKernelException.getTraceDepth();
		
		ThreadInfo[] threadInfo=threadBean.getThreadInfo(ids, maxDepth);	
		
		//get a measure of cpuTime
		long[] cpuTimes=new long[ids.length];
		if (threadBean.isThreadCpuTimeSupported())
		{	boolean cpuTimeEnabled=threadBean.isThreadCpuTimeEnabled();
			threadBean.setThreadCpuTimeEnabled(true);
			final long DURATION=500;
			
			for (int i=0; i<ids.length; i++)
			{	cpuTimes[i]=threadBean.getThreadCpuTime(ids[i]);
			}
			Thread.sleep(DURATION);
			for (int i=0; i<ids.length; i++)
			{	long t=threadBean.getThreadCpuTime(ids[i]);
				if (cpuTimes[i]>=0 && t>=0)
				{	cpuTimes[i]=(t-cpuTimes[i])/(DURATION);
				}
				else
				{	cpuTimes[i]=0;
				}
			}
			threadBean.setThreadCpuTimeEnabled(cpuTimeEnabled);
		}
		
		//find deadlocks
		long[] deadlockedIds=threadBean.findMonitorDeadlockedThreads();		
		
		
		Map<Long,RootStatus> rootStatus = new HashMap<Long,RootStatus>();
		long now=System.currentTimeMillis();
		HDSBuilder b=new HDSBuilder();
		b.pushNode("k");

		//locked endpoints
		b.pushNode("locked-endpoints");
		for (Iterator i=lockedEndpoints.keyIterator(); i.hasNext(); )
		{	Object endpoint = i.next();
			b.pushNode("endpoint");
			b.addNode("class", endpoint.getClass().getName());
			b.pushNode("queued");
			List<RequestState> lockedStates=lockedEndpoints.get(endpoint);
			for (RequestState rs: lockedStates)
			{	long id=0;
				IRequestState root=rs;
				while(root.getParent()!=null)
				{	root=root.getParent();
				}
				id=((RootState)root).getId();
				b.pushNode("state");
				b.addNode("root-id",id);
				rootStatus.put(id, RootStatus.BUSY_ENDPOINT);
				b.addNode("request",RequestFactory.toShortString(rs.getRequestIn()));
				b.addNode("active",now-rs.getTimestamp());
				b.popNode();
			}
			b.popNode();
			b.popNode();
		}
		b.popNode();
		
		//queued endpoints
		b.pushNode("queue");
		for (RequestState rs: queue)
		{	
			IRequestState root=rs;
			IRequestState parent;
			while( (parent=root.getParent())!=null)
			{	root=parent;
			}
			long id=((RootState)root).getId();
			rootStatus.put(id, RootStatus.QUEUED);
			b.pushNode("state");
			b.addNode("root-id",id);
			b.addNode("request",RequestFactory.toShortString(rs.getRequestIn()));
			long ts=rs.getTimestamp();
			if (ts==0)
			{	if (rs.getParent()!=null)
				{	ts=((RequestState)rs.getParent()).getTimestamp();
				}
				else
				{	ts=now;
				}
			}
			b.addNode("active",now-ts);			
			b.addNode("priority",RequestFactory.priorityString(rs.getRequestIn().getPriority()));			
			b.addNode("state",((Object)rs.getState()).toString());			
			b.popNode();
		}
		b.popNode();
		
		//threads
		
		NumberFormat df=NumberFormat.getInstance(Locale.US);
		df.setMaximumFractionDigits(2);
		df.setMinimumFractionDigits(2);
		
		b.pushNode("threads");
		int idleCount=0;
		for (int j=0; j<ids.length; j++)
		{	ThreadInfo ti=threadInfo[j];
			if (ti==null) continue;
			
			StackTraceElement[] stes=ti.getStackTrace();
			RequestState last=lastStateOnThread.get(ti.getThreadId());
			
			b.pushNode("thread");
			long id=ti.getThreadId();
			b.addNode("id",id);
			b.addNode("name",ti.getThreadName());
			b.addNode("cpu",df.format(((double)cpuTimes[j])/10000.0));
			b.addNode("isKernel",tidList.contains(id));
			boolean deadlocked=false;
			if (deadlockedIds!=null)
			{	for (int i=0; i<deadlockedIds.length; i++)
				{	if (deadlockedIds[i]==id)
					{	deadlocked=true;
						break;
					}
				}
			}
			
			String state;
			if (deadlocked)
			{	state="DEADLOCK";
			}
			else if (stes.length>5 && stes[5].getClassName().endsWith("PriorityRequestTable"))
			{	state="IDLE";
				idleCount++;
			}
			else if (stes.length>6 && stes[6].getClassName().endsWith("SynchronousRequestee"))
			{	state="SYNC_JOIN";
			}
			else if (stes.length>0 && (stes[0].getClassName().startsWith("java.net.") || stes[0].getClassName().startsWith("java.io.")  || stes[0].getClassName().startsWith("java.nio.")))
			{	state="IO_WAIT";
			}
			else
			{	state=ti.getThreadState().name();
			}
			b.addNode("state",state);
			if (state.equals("IDLE"))
			{	last=null;
			}
			
			//lastStateOnThread
			String lock=ti.getLockName();
			if (lock!=null)
			{	b.addNode("lock",lock);
				b.addNode("lock-owner",ti.getLockOwnerName());
			}
			if (last!=null)
			{	b.addNode("root",last.getRoot().getId());
				b.addNode("request",RequestFactory.toShortString(last.getRequestIn()));
			}
			b.pushNode("stack");
			for (int i=0; i<stes.length; i++)
			{	String frame=stes[i].getClassName()+"."+stes[i].getMethodName()+"() line:"+Integer.toString(stes[i].getLineNumber());
				b.addNode("frame", frame);
			}
			b.popNode();
			
			b.popNode();
			
			
			//detected root status problems
			if (last!=null)
			{	
				b.addNode("active",now-last.getTimestamp());
				
				Long rootId=last.getRoot().getId();
			
				RootStatus rs=rootStatus.get(rootId);
				long active=now-last.getTimestamp();
				if ( active>runawayPeriod)
				{
					Object lt=last.getRequestIn().getHeaders().getValue(INKFRequest.HEADER_LIVENESS);
					if (!(lt instanceof Long) || ((Long) lt).longValue()<active)
					{	
						if (state.equals("RUNNABLE"))
						{	rootStatus.put(rootId,RootStatus.RUNAWAY);
						}
						else if (state.contains("WAITING") || state.equals("BLOCKED"))
						{	rootStatus.put(rootId,RootStatus.LONG_WAIT);
						}
						else if (state.equals("IO_WAIT"))
						{	rootStatus.put(rootId,RootStatus.LONG_IO_WAIT);
						}
						else
						{	//System.out.println(state);
						}
					}
				}
				
				if (deadlocked)
				{	rootStatus.put(rootId, RootStatus.DEADLOCKED);
				}
			}
		}
		b.popNode();		
		
		//detected starvation && responselessness
		for (RootState root: rootStates)
		{	Long rootId=root.getId();
			Set<Long> threadsForRoot = new HashSet();
			recurseThreadsForRoot(root, threadsForRoot);
			int syncCount=0;
			int rootIdleCount=0;
			for(Long threadId: threadsForRoot)
			{	ThreadInfo ti=null;
				for (int i=0; i<threadInfo.length; i++)
				{	ThreadInfo ti2=threadInfo[i];
					if (ti2==null) continue;
					if (ti2.getThreadId()==threadId)
					{	ti=ti2;
						break;
					}
				}
				if (ti!=null)
				{	StackTraceElement[] stes=ti.getStackTrace();
					if (stes.length>6 && stes[6].getClassName().endsWith("SynchronousRequestee"))
					{	syncCount++;
					}
					if (stes.length>4 && stes[4].getClassName().endsWith("PriorityRequestTable"))
					{	rootIdleCount++;
					}
				}
			}
			if (syncCount==threadsForRoot.size())
			{	long activity;
				RequestState last = lastStateOnRoot.get(root);
				if (last!=null)
				{	activity=now-last.getTimestamp();
				}
				else
				{	activity=Long.MAX_VALUE;
				}
			
				if (idleCount==0)
				{	if (activity>runawayPeriod)
					{	rootStatus.put(rootId, RootStatus.THREAD_STARVED);
					}
				}
				else if (rootStatus.get(rootId)==RootStatus.BUSY_ENDPOINT) 
				{	//leave it as busy
				}
				else if (rootStatus.get(rootId)==null) //a synchronous parent is waiting but no threads active or requests queued
				{	
					
					if (activity>runawayPeriod)
					{	//System.out.println("responseless 1 "+ root.getRequestIn()+" "+rootStatus.get(rootId));
						rootStatus.put(rootId, RootStatus.RESPONSELESS);
					}
				}
			}
			else if ((threadsForRoot.size()-rootIdleCount)==0 && rootStatus.get(rootId)!=RootStatus.QUEUED)
			{	long activity;
				RequestState last = lastStateOnRoot.get(root);
				if (last!=null)
				{	activity=now-last.getTimestamp();
				}
				else
				{	activity=Long.MAX_VALUE;
				}
				if (activity>runawayPeriod)
				{	//System.out.println("responseless 2");
					rootStatus.put(rootId, RootStatus.RESPONSELESS);
				}
			}
		}
		
		//output root requests
		b.pushNode("roots");
		for (RootState root: rootStates)
		{	Long rootId=root.getId();
			b.pushNode("state");
			
			b.addNode("root-id",rootId);
			b.addNode("request",RequestFactory.toShortString(root.getRequestIn()));
			long duration=now-root.getTimestamp();
			b.addNode("duration",duration);
			RootStatus status = rootStatus.get(rootId);
			if (status==null || status==RootStatus.QUEUED)
			{	status=RootStatus.OK;
			}
			b.addNode("status", status);
			b.addNode("statusString", status.toString());
			
			RequestState last = lastStateOnRoot.get(root);
			if (last!=null)
			{	long activity=now-last.getTimestamp();
				b.addNode("active",activity);
			}
			
			b.popNode();
		}
		b.popNode();			
		
		b.popNode();
		return b.getRoot();
	}
	
	/** Update aLastStateOnThread
	 * @param aState
	 * @param aLastStateOnThread
	 * @return latest state on or below aState
	 */
	private RequestState recurseStateForThreads(RequestState aState, Map<Long,RequestState> aLastStateOnThread)
	{	RequestState result=null;
		Thread t=aState.getExecutionThread();
		if (t!=null)
		{	Long id=t.getId();
			RequestState latest=aLastStateOnThread.get(id);
			if (latest!=null)
			{	if (latest.getTimestamp()<=aState.getTimestamp())
				{	aLastStateOnThread.put(id, aState);
					result=aState;
				}
			}
			else
			{	aLastStateOnThread.put(id, aState);
				result=aState;
			}
		}
		List<RequestState> substates=aState.getSubStates();
		for (RequestState substate: substates)
		{	RequestState rs=recurseStateForThreads(substate,aLastStateOnThread);
			if (rs!=null)
			{	if (result!=null)
				{	if (result.getTimestamp()<=rs.getTimestamp())
					{	result=rs;
					}
				}
				else
				{	result=rs;
				}
			}
		}
		return result;
	}
	
	private void recurseThreadsForRoot(RequestState aState, Set<Long> aThreadsForRoot)
	{
		Thread t=aState.getExecutionThread();
		if (t!=null)
		{	Long tid=t.getId();
			aThreadsForRoot.add(tid);
		}
		for (RequestState s: aState.getSubStates())
		{	recurseThreadsForRoot(s,aThreadsForRoot);
		}
	}
	
	public enum RootStatus { OK, QUEUED, BUSY_ENDPOINT, RUNAWAY, DEADLOCKED, RESPONSELESS, LONG_WAIT, LONG_IO_WAIT, THREAD_STARVED }

	
	private IHDSNode getKernelSpaces(INKFRequestContext aContext) throws Exception
	{	HDSBuilder b = new HDSBuilder();
		b.pushNode("spaces");
		IRepDeployedModules rep=(IRepDeployedModules)aContext.source("active:moduleList",IRepDeployedModules.class);
		for (Iterator<ISpaceWithIdentity> i=getKernel().getSpaces(); i.hasNext(); )
		{	ISpaceWithIdentity space=i.next();
			b.pushNode("space");
			b.addNode("id",space.getIdentifier());
			b.addNode("version",space.getVersion());
			String spaceHash=rep.hashForComponent(space);
			b.addNode("hash", spaceHash);
			IMetaRepresentation meta=rep.metadataForHash(spaceHash);
			b.addNode("name", meta.getName());
			b.addNode("desc", meta.getDescription());
			b.popNode();
		}
		return b.getRoot();
	}
	
	private IHDSNode getEnvironment(INKFRequestContext aContext) throws Exception
	{	HDSBuilder b = new HDSBuilder();
		b.pushNode("environment");
		
		Properties p = System.getProperties();
		b.pushNode("system-properties");
		for (Entry e : p.entrySet())
		{	b.pushNode("property");
			b.addNode("name",e.getKey());
			b.addNode("value",e.getValue());
			b.popNode();
		}
		b.popNode();

		Runtime r=Runtime.getRuntime();
		b.addNode("processors",Integer.toString(r.availableProcessors()));
		b.addNode("max-memory",Long.toString(Runtime.getRuntime().maxMemory()));
		
		InetAddress localMachine = InetAddress.getLocalHost();
		b.addNode("hostname",localMachine.getHostName());
		b.addNode("ip",localMachine.getHostAddress());
		
		b.pushNode("memory-pools");
		{
			List<MemoryPoolMXBean> pools= ManagementFactory.getMemoryPoolMXBeans();
			Map<String,MemoryPoolMXBean> poolsByName = new HashMap<String,MemoryPoolMXBean>();
			for (MemoryPoolMXBean pool : pools)
			{	poolsByName.put(pool.getName(), pool);
			}
			final String[] poolOrder = new String[] { "PS Eden Space", "PS Survivor Space", "PS Old Gen", "Code Cache", "PS Perm Gen"};
			int t=0;
			for (int i=0; i<poolOrder.length; i++)
			{	if (poolsByName.containsKey(poolOrder[i])) t++;
			}
			if (t==poolOrder.length)
			{	for (int i=0; i<poolOrder.length; i++)
				{	outputMemoryUsage(b,poolOrder[i],poolsByName.get(poolOrder[i]).getUsage());
				}
			}
			else
			{	for (MemoryPoolMXBean pool : pools)
				{	outputMemoryUsage(b,pool.getName(),pool.getUsage());
				}
			}
		}
		b.popNode();
		
		b.pushNode("garbage-collectors");
		List<GarbageCollectorMXBean> gcs=ManagementFactory.getGarbageCollectorMXBeans();
		for (GarbageCollectorMXBean bean: gcs)
		{	b.pushNode("collector");
			b.addNode("name",bean.getName());
			b.addNode("count",bean.getCollectionCount());
			b.addNode("time",bean.getCollectionTime());
			b.popNode();
		}
		b.popNode();
		
		
		ClassLoader cl=aContext.getKernelContext().getKernel().getClass().getClassLoader();
		if (cl instanceof URLClassLoader)
		{	//should always be
			URLClassLoader ucl = (URLClassLoader)cl;
			URL[] bootURLs=ucl.getURLs();
			
			b.pushNode("boot-classpath");
			for (URL url : bootURLs)
			{	b.pushNode("element");
				String urlString=url.toString();
				b.addNode("url",urlString);
				if (urlString.endsWith(".jar"))
				{	
					try
					{	URL jarURL=new URL("jar:"+urlString+"!/");
						JarURLConnection con=(JarURLConnection)jarURL.openConnection();
						Manifest mf=con.getManifest();
						String title=getManifestValue(mf,"Implementation-Title");
						String version=getManifestValue(mf,"Implementation-Version");
						if (title!=null) b.addNode("title",title);
						if (version!=null) b.addNode("version",version);
					}
					catch (Exception e)
					{ ; }
				}
				b.popNode();
			}
			b.popNode();
		}
		
		b.pushNode("modules");
		ModuleManager mm=ModuleManager.getSingleton();
		Map<URI,ModuleSourceWithLevel> rm=mm.getRegisteredModules();
		for (IModule m: mm.getModules())
		{	b.pushNode("module");
			IModuleMeta meta=m.getMeta();
			b.addNode("identifier",meta.getIdentifier());
			b.addNode("version",meta.getVersion());
			ModuleSourceWithLevel mswl=rm.get(meta.getSource());
			if (mswl!=null)
			{	b.addNode("runlevel",mswl.getLevel());
			}
			b.popNode();
		}
		b.popNode();
		
		
		b.addNode("runlevel",ModuleManager.getSingleton().getRunLevel());
		
		try
	    {	if (! File.separator.equals("/")) throw new Exception();
			INKFRequest req=aContext.createRequest("active:exec");
			req.addArgumentByValue("command","uptime");
			req.setRepresentationClass(String.class);
			String uptimeString=(String)aContext.issueRequest(req);
			b.addNode("osuptime",uptimeString);
	    } catch (Exception e)
	    {	b.addNode("osuptime","OS uptime not available");
		}
	    
	    
		String pid="PID not available";
		try
		{
		    List<String> commands = new ArrayList<String>();
			commands.add("/bin/bash");
			commands.add("-c");
			commands.add("echo $PPID");
			ProcessBuilder pb = new ProcessBuilder(commands);
			Process pr = pb.start();
			pr.waitFor();
			if (pr.exitValue() == 0)
			{	BufferedReader outReader = new BufferedReader(new InputStreamReader(pr.getInputStream()));
				pid= outReader.readLine().trim();
				
				commands.clear();
				commands.add("ps");
				commands.add("o");
				commands.add("rss vsz pmem");
				commands.add("hp");
				commands.add(pid);
				pr = pb.start();
				pr.waitFor();
				if (pr.exitValue() == 0)
				{	outReader = new BufferedReader(new InputStreamReader(pr.getInputStream()));
					String mem= outReader.readLine().trim();
					if (mem.contains("RSS"))
					{	//mac ps doesn't support h option to remove header
						mem= outReader.readLine().trim();
					}
					StringTokenizer st=new StringTokenizer(mem," ");
					long resident=Long.parseLong(st.nextToken())*1024L;
					long virtual=Long.parseLong(st.nextToken())*1024L;
					long percent=(long)(10.0f*Float.parseFloat(st.nextToken()));
					long physical= resident*1000/percent;
					b.addNode("resident",resident);
					b.addNode("virtual",virtual);
					b.addNode("physical",physical);
				}
				
			}
		}
		catch (Exception e)
		{
		}

		b.addNode("pid",pid);
		
		//request/cache/memory summary
		IHDSNode stats=aContext.source("netkernel:/memory",IHDSNode.class);
		long totalReqInPeriod=0;
		long peakMemory=0;
		long baselineMemory=0;
		long cacheSize=0;
		long t0=0;
		long t1=0;
		for (IHDSNode stat : stats.getNodes("/memory/stat"))
		{	t0++;
		
			long v=(Long)stat.getFirstValue("requests");
			totalReqInPeriod+=v;
			
			v =(Long)stat.getFirstValue("peak");
			peakMemory+=v;
			if (v>0)
			{	t1++;
			}
			v =(Long)stat.getFirstValue("baseline");
			baselineMemory+=v;
			v =(Long)stat.getFirstValue("cacheSize");
			cacheSize+=v;
			
			
		}
		if (t1==0) t1=1;
		if (t0==0) t0=1;
		peakMemory/=t1;
		baselineMemory/=t1;
		cacheSize/=t1;
		long period=(Long)stats.getFirstValue("/memory/totalDurationSeconds");
		period=period*t1/t0;
		b.addNode("totalRequests",RootState.getNextId());
		
		b.addNode("statPeriod", period);
		b.addNode("requestsInPeriod",totalReqInPeriod);
		b.addNode("requestRateInPeriod",((float)totalReqInPeriod)/((float)period));
		b.addNode("avgPeakMemory",peakMemory*1024);
		b.addNode("avgBaselineMemory",baselineMemory*1024);
		b.addNode("avgCacheSize",cacheSize);
		
		//OS Environment Properties
		Map<String, String> env = System.getenv();
		b.pushNode("os-properties");
		for (Entry e : env.entrySet()) {
			b.pushNode("property");
				b.addNode("name",e.getKey());
				b.addNode("value",e.getValue());
			b.popNode();
		}
		b.popNode();
		
		return b.getRoot();
	}
	
	private String getManifestValue(Manifest aManifest, String aName)
	{	String result=null;
		for (Entry<String,Attributes> e : aManifest.getEntries().entrySet())
		{	Attributes a=e.getValue();
			for (Object k : a.keySet())
			{	if (k.toString().equals(aName))
				{	result=a.getValue(k.toString());
					break;
				}
			}
		}
		return result;
	}
	
	private static 	void	outputMemoryUsage(HDSBuilder b, String aName, MemoryUsage aUsage)
	{	b.pushNode("pool");
		b.addNode("name",aName);
		b.addNode("committed",aUsage.getCommitted());
		b.addNode("init",aUsage.getInit());
		b.addNode("max",aUsage.getMax());
		b.addNode("used",aUsage.getUsed());
		b.popNode();
	}
	
	private IHDSNode getDeployment(INKFRequestContext aContext) throws Exception
	{
		IConfiguration config=aContext.getKernelContext().getKernel().getConfiguration();
		String installPathS=config.getString("netkernel.install.path");
		String initModulesS=config.getString("netkernel.init.modules");
		String stemModulesS=config.getString("netkernel.init.stem.modules");

		URI installPath=URI.create(installPathS);
		URI initModules=installPath.resolve(initModulesS);
		
		HDSBuilder b = new HDSBuilder();
		b.pushNode("modules");
		List<IModule> modules=ModuleManager.getSingleton().getModules();
		Map<URI,ModuleSourceWithLevel> registeredModules=ModuleManager.getSingleton().getRegisteredModules();
		
		//scan stem modules
		InputStream is = IKernel.class.getClassLoader().getResourceAsStream(stemModulesS);
		BufferedReader r = new BufferedReader(new InputStreamReader(is));
		String line;
		while ((line=r.readLine())!=null)
		{	line=line.trim();
			if (!line.startsWith("#") && line.length()>0)
			{	URI source = installPath.resolve(line);
				IModule m = getModuleWithSource(source,modules);
				if (m!=null)
				{
					b.pushNode("module");
					b.addNode("id",m.getMeta().getIdentifier());
					b.addNode("runlevel","1");
					b.addNode("src",source.toString());
					b.addNode("jar",!source.toString().endsWith("/"));
					b.addNode("enabled",true);
					b.popNode();
				}
			}
		}
		
		//scan modules.xml
		IHDSNode modulesXML = aContext.source(initModules.toString(),IHDSNode.class);
		for (IHDSNode n : modulesXML.getNodes("/modules/*"))
		{
			line = n.getValue().toString().trim();
			b.pushNode("module");
			URI source = installPath.resolve(line);
			IModule m = getModuleWithSource(source,modules);
			if (m!=null)
			{	IModuleMeta meta=m.getMeta();
				b.addNode("id",meta.getIdentifier());
				b.addNode("version",meta.getVersion());
				b.addNode("name",meta.getName());
				b.addNode("desc",meta.getDescription());
			}
			b.addNode("src",source.toString());
			
			String runlevel = (String)n.getFirstValue("@runlevel");
			if (runlevel==null)
			{	runlevel = "*";
			}
			b.addNode("runlevel",runlevel);
			b.addNode("jar",!source.toString().endsWith("/"));
			String name=n.getName();
			b.addNode("enabled",name.equals("module"));
			b.popNode();
		}
		
		return b.getRoot();
		
	}
	
	private static IModule getModuleWithSource(URI aSrc, List<IModule> aModules)
	{	IModule result=null;
		for (IModule m: aModules)
		{	if (m.getSource().equals(aSrc))
			{	result=m;
				break;
			}
		}
		return result;
	}


}
