package org.netkernel.ext.system.accessor;

import java.io.File;
import java.net.JarURLConnection;
import java.net.URI;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Enumeration;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import org.netkernel.container.IKernel;
import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.layer0.boot.BootUtils;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.representation.impl.HDSPredicateFactory;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.util.RequestBuilder;
import org.netkernel.layer0.util.Utils;
import org.netkernel.module.standard.StandardSpace;
import org.netkernel.module.standard.builtin.StaticResourceEndpoint;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpaceWithIdentity;
import org.netkernel.urii.impl.Version;

public class SpaceStaticResourceListAccessor extends StandardAccessorImpl
{
	public void onSource(INKFRequestContext aContext) throws Exception
	{
		String spaceIdentifier=aContext.getThisRequest().getArgumentValue("space");
		String versionString=aContext.getThisRequest().getArgumentValue("version");
		Version version=null;
		if (versionString!=null)
		{	version=new Version(versionString);
		}
		IKernel k=aContext.getKernelContext().getKernel();
		ISpaceWithIdentity space=k.getSpace(new SimpleIdentifierImpl(spaceIdentifier), version, version);
		if (space==null)
		{	throw new NKFException("Space not found");
		}
		if (!(space instanceof StandardSpace))
		{	throw new NKFException ("Space not a standard module space");
		}

		StandardSpace ss=(StandardSpace)space;
		String source=ss.getOwningModule().getSource().toString();
		source=BootUtils.fixURIString(source);
		HDSBuilder b = new HDSBuilder();
		b.pushNode("dir");
		if (source.startsWith("file:"))
		{	File f = new File(new URI(source));
			StringBuilder path = new StringBuilder(256);
			path.append("res:/");
			recurseDir(f,b,path );
			INKFResponse response=aContext.createResponseFrom(b.getRoot());
			long pollPeriod=aContext.getKernelContext().getKernel().getConfiguration().getLong("netkernel.poll", 5000L);
			response.setExpiry(INKFResponse.EXPIRY_CONSTANT, System.currentTimeMillis()+pollPeriod);
		}
		else if (source.startsWith("jar:"))
		{	scanJar(source,b);
			aContext.createResponseFrom(b.getRoot());
		}
		else
		{	throw new NKFException ("unspported module source "+source);
		}
		
		
		
		
		/*
		URLClassLoader ucl=null;
		ucl.getURLs();
		
		//space.getClassLoader().
		
		IRepDeployedModules modules=aContext.source("active:moduleList",IRepDeployedModules.class);
		String hash=modules.hashForComponent(space);
		ISpaceMeta meta=(ISpaceMeta)modules.metadataForHash(hash);
		ISpaceElements elements=meta.getElements();
		for (int i=0; i<elements.size(); i++)
		{	String id=elements.getIdentifier(i).toString();
			String elementHash=hash+"/"+id;
			IMetaRepresentation elementMeta=(IMetaRepresentation)modules.metadataForHash(elementHash);
			if (elementMeta instanceof IEndpointMeta)
			{	IEndpointMeta epMeta=(IEndpointMeta)elementMeta;
				epMeta.getIdentifierGrammar().
			}
			
			IEndpoint physicalEp=elements.getPhysicalEndpoint(i);
			if (physicalEp instanceof StaticResourceEndpoint)
			{
		}
		*/
	}
	
	private void scanJar(String aSource, HDSBuilder aBuilder) throws Exception
	{
		long time = System.currentTimeMillis();
		URL url = new URL(aSource);
		JarURLConnection jarCon = (JarURLConnection)url.openConnection();
		JarFile jf = jarCon.getJarFile();
		StringBuilder path = new StringBuilder(256);
		path.append("res:/");
		for (Enumeration e = jf.entries(); e.hasMoreElements(); )
		{	JarEntry je = (JarEntry)e.nextElement();
			if (!je.isDirectory())
			{	path.setLength(5);
				String name=je.getName();
				String[] pe=Utils.splitString(name, "/");
				IHDSNode c=aBuilder.getRoot();
				for (int i=0; i<pe.length-1; i++)
				{
					String dirName=pe[i];
					path.append(dirName);
					path.append("/");
					IHDSNodeList nl=c.getNodes("dir").filter(HDSPredicateFactory.namedChildStringEquals("@name", dirName));
					if (nl.size()>0)
					{	c=nl.get(0);
					}
					else
					{	aBuilder.setCursor(c);
						aBuilder.pushNode("dir");
						aBuilder.addNode("@name", dirName);
						aBuilder.addNode("@id", path.toString());
						c=aBuilder.getCursor();
					}
				}
				aBuilder.setCursor(c);
				aBuilder.pushNode("res");
				String resName=pe[pe.length-1];
				aBuilder.addNode("@name", resName);
				path.append(resName);
				aBuilder.addNode("@id", path.toString());
			}
			else
			{	//can we rely on directories being before entries? not sure
			}
		}
		
	}
	
	private void recurseDir(File aFile, HDSBuilder aBuilder, StringBuilder aPath)
	{
		long time = System.currentTimeMillis();
		File[] files = aFile.listFiles();
		int pathLength = aPath.length();
		for (int i=0; i<files.length; i++)
		{	File f = files[i];
			String name = f.getName();
			aPath.setLength(pathLength);
			Utils.appendUnreservedURIChar(aPath,name);
			if (f.isHidden())
			{
			}
			else if (f.isDirectory())
			{	aPath.append('/');
				aBuilder.pushNode("dir");
				aBuilder.addNode("@id",aPath.toString());
				aBuilder.addNode("@name",name);
				recurseDir(f,aBuilder,aPath);
				aBuilder.popNode();
			}
			else
			{	aBuilder.pushNode("res");
				aBuilder.addNode("@id",aPath.toString());
				aBuilder.addNode("@name",name);
				aBuilder.popNode();
			}
		}
		
	}
	
}
